#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
from abc import ABC, ABCMeta, abstractmethod
from typing import Any, TypeVar, Generic, Callable

from ..pdi.base3_component import ConsistComponent, RouteComponent
from ..pdi.pdi_req import PdiReq

from ..protocol.command_req import CommandReq
from ..protocol.multibyte.multibyte_constants import TMCC2EffectsControl
from ..protocol.constants import CommandScope
from ..utils.text_utils import title

log = logging.getLogger(__name__)

BASE_TO_TMCC_SMOKE_MAP = {
    0: TMCC2EffectsControl.SMOKE_OFF,
    1: TMCC2EffectsControl.SMOKE_LOW,
    2: TMCC2EffectsControl.SMOKE_MEDIUM,
    3: TMCC2EffectsControl.SMOKE_HIGH,
}

TMCC_TO_BASE_SMOKE_MAP = {v: k for k, v in BASE_TO_TMCC_SMOKE_MAP.items()}


def default_from_func(t: bytes) -> int:
    return int.from_bytes(t, byteorder="little")


def default_to_func(t: int) -> bytes:
    return t.to_bytes(1, byteorder="little")


class CompDataHandler:
    """
    Helper class to read and write configuration data from and to bytes.
    Lionel component state is read from the Base 3 using the BASE_MEMORY,
    D4_ENGINE, and D4_TRAIN pdi commands. State is returned as a byte string.
    The CompDataHandler class defines the byte address where specific state
    is stored and allows lambda functions to convert the raw byte string to
    Python types as well as to write Python types to the byte string.
    """

    def __init__(
        self,
        field: str,
        length: int = 1,
        from_bytes: Callable = default_from_func,
        to_bytes: Callable = default_to_func,
        d4_only: bool = False,
    ) -> None:
        self.field: str = field
        self.length: int = length
        self.from_bytes: Callable = from_bytes
        self.to_bytes: Callable = to_bytes
        self._d4_only: bool = d4_only

    @property
    def is_d4_only(self) -> bool:
        return self._d4_only


class UpdatePkg:
    def __init__(self, field: str, offset: int, length: int, data_bytes: bytes) -> None:
        self.field: str = field
        self.offset: int = offset
        self.length: int = length
        self.data_bytes: bytes = data_bytes

    def __repr__(self) -> str:
        return f"{self.field}: Address: {hex(self.offset)} Length: {self.length} data: {self.data_bytes.hex()}"


#
# Base 3 memory locations where device state is stored. When commands are issued that change
# device characteristics, like engine speed or momentum, these changes must be explicitly
# written to the Base 3 so it can update other attached controllers and software.
#
BASE_MEMORY_ENGINE_READ_MAP = {
    0xB8: CompDataHandler(
        "_tmcc_id",
        4,
        lambda t: int(PdiReq.decode_text(t)),
        lambda t: PdiReq.encode_text(str(t).zfill(4), 4),
        True,
    ),
    0x00: CompDataHandler("_prev_link"),
    0x01: CompDataHandler("_next_link"),
    0x02: CompDataHandler("_unk_2"),
    0x03: CompDataHandler("_unk_3"),
    0x04: CompDataHandler(
        "_bt_id",
        2,
        lambda t: int.from_bytes(t[0:2], byteorder="little"),
        lambda t: t.to_bytes(2, byteorder="little"),
    ),
    0x06: CompDataHandler("_unk_6"),
    0x07: CompDataHandler("_speed"),
    0x08: CompDataHandler("_target_speed"),
    0x09: CompDataHandler("_train_brake"),
    0x0A: CompDataHandler("_unk_a"),
    0x0B: CompDataHandler("_unk_b"),
    0x0C: CompDataHandler("_rpm_labor"),
    0x0D: CompDataHandler("_fuel_level"),
    0x0E: CompDataHandler("_water_level"),
    0x0F: CompDataHandler("_unk_f"),
    0x11: CompDataHandler("_unk_11"),
    0x12: CompDataHandler("_unk_12"),
    0x13: CompDataHandler("_unk_13"),
    0x17: CompDataHandler("_unk_17"),
    0x18: CompDataHandler("_momentum"),
    0x1E: CompDataHandler("_road_name_len"),
    0x1F: CompDataHandler(
        "_road_name",
        31,
        lambda t: PdiReq.decode_text(t),
        lambda t: PdiReq.encode_text(t, 31),
    ),
    0x3E: CompDataHandler("_road_number_len"),
    0x3F: CompDataHandler(
        "_road_number",
        4,
        lambda t: PdiReq.decode_text(t),
        lambda t: PdiReq.encode_text(t, 4),
    ),
    0x43: CompDataHandler("_engine_type"),
    0x44: CompDataHandler("_control_type"),
    0x45: CompDataHandler("_sound_type"),
    0x46: CompDataHandler("_engine_class"),
    0x59: CompDataHandler("_tsdb_left"),
    0x5B: CompDataHandler("_tsdb_right"),
    0x5C: CompDataHandler("_unk_5c"),
    0x5E: CompDataHandler("_unk_5e"),
    0x68: CompDataHandler("_unk_68"),
    0x69: CompDataHandler("_smoke"),
    0x6A: CompDataHandler("_speed_limit"),
    0x6B: CompDataHandler("_max_speed"),
    0xBC: CompDataHandler(
        "_timestamp",
        4,
        lambda t: int.from_bytes(t[0:4], byteorder="little"),
        lambda t: t.to_bytes(4, byteorder="little"),
        True,
    ),
}
# build an inverse map from the token name to the address
FIELD_TO_ADDR_ENGINE_MAP = {v.field[1:]: k for k, v in BASE_MEMORY_ENGINE_READ_MAP.items()}

BASE_MEMORY_TRAIN_READ_MAP = {
    0x6F: CompDataHandler("_consist_flags"),
    0x70: CompDataHandler(
        "_consist_comps",
        32,
        lambda t: ConsistComponent.from_bytes(t),
        lambda t: ConsistComponent.to_bytes(t),
    ),
}
BASE_MEMORY_TRAIN_READ_MAP.update(BASE_MEMORY_ENGINE_READ_MAP)

BASE_MEMORY_ACC_READ_MAP = {
    0x00: CompDataHandler("_prev_link"),
    0x01: CompDataHandler("_next_link"),
    0x1E: CompDataHandler("_device_code"),
    0x1F: CompDataHandler(
        "_road_name",
        31,
        lambda t: PdiReq.decode_text(t),
        lambda t: PdiReq.encode_text(t, 31),
    ),
    0x3E: CompDataHandler("_road_number_len"),
    0x3F: CompDataHandler(
        "_road_number",
        4,
        lambda t: PdiReq.decode_text(t),
        lambda t: PdiReq.encode_text(t, 4),
    ),
}

BASE_MEMORY_SWITCH_READ_MAP = {
    0x00: CompDataHandler("_prev_link"),
    0x01: CompDataHandler("_next_link"),
    0x05: CompDataHandler(
        "_road_name",
        31,
        lambda t: PdiReq.decode_text(t),
        lambda t: PdiReq.encode_text(t, 31),
    ),
    0x24: CompDataHandler("_road_number_len"),
    0x25: CompDataHandler(
        "_road_number",
        4,
        lambda t: PdiReq.decode_text(t),
        lambda t: PdiReq.encode_text(t, 4),
    ),
}

BASE_MEMORY_ROUTE_READ_MAP = BASE_MEMORY_SWITCH_READ_MAP.copy()
BASE_MEMORY_ROUTE_READ_MAP.update(
    {
        0x60: CompDataHandler(
            "_components",
            32,
            lambda t: RouteComponent.from_bytes(t),
            lambda t: RouteComponent.to_bytes(t),
        ),
    }
)

SCOPE_TO_COMP_MAP = {
    CommandScope.ENGINE: BASE_MEMORY_ENGINE_READ_MAP,
    CommandScope.TRAIN: BASE_MEMORY_TRAIN_READ_MAP,
    CommandScope.ACC: BASE_MEMORY_ACC_READ_MAP,
    CommandScope.SWITCH: BASE_MEMORY_SWITCH_READ_MAP,
    CommandScope.ROUTE: BASE_MEMORY_ROUTE_READ_MAP,
}

#
# Map of Base 3 Command Requests to the corresponding Base 3 state updates
# that must be made.
#
REQUEST_TO_UPDATES_MAP = {
    "ABSOLUTE_SPEED": [
        ("speed",),
        ("target_speed",),
    ],
    "DIESEL_RPM": [("rpm",)],
    "ENGINE_LABOR": [("labor",)],
    "MOMENTUM": [("momentum",)],
    "MOMENTUM_HIGH": [("momentum", lambda t: 127)],
    "MOMENTUM_LOW": [("momentum", lambda t: 0)],
    "MOMENTUM_MEDIUM": [("momentum", lambda t: 63)],
    "RESET": [
        ("speed", lambda x: 0),
        ("target_speed", lambda x: 0),
        ("rpm_labor", lambda x: 0),
    ],
    "SHUTDOWN_DELAYED": [("rpm_labor", lambda x: 0)],
    "SHUTDOWN_IMMEDIATE": [("rpm_labor", lambda x: 0)],
    "SMOKE_HIGH": [("smoke", lambda t: 3)],
    "SMOKE_LOW": [("smoke", lambda t: 1)],
    "SMOKE_MEDIUM": [("smoke", lambda t: 2)],
    "SMOKE_OFF": [("smoke", lambda t: 0)],
    "SMOKE_ON": [("smoke", lambda t: 1)],
    "STOP_IMMEDIATE": [
        ("speed", lambda x: 0),
        ("target_speed", lambda x: 0),
        ("rpm_labor", lambda x: 0),
    ],
    "TRAIN_BRAKE": [("train_brake",)],
}

CONVERSIONS: dict[str, tuple[Callable, Callable]] = {
    "train_brake": (lambda x: min(round(x * 0.4667), 7), lambda x: min(round(x * 2.143), 15)),
    "momentum": (lambda x: min(round(x * 0.05512), 7), lambda x: min(round(x * 18.14), 127)),
    "smoke": (
        lambda x: BASE_TO_TMCC_SMOKE_MAP.get(x, TMCC2EffectsControl.SMOKE_OFF),
        lambda x: TMCC_TO_BASE_SMOKE_MAP.get(x, 0),
    ),
    "rpm": (lambda x: x & 0b111, lambda x: x & 0b111),
    "labor": (
        lambda x: (x >> 3) + 12 if (x >> 3) <= 19 else (x >> 3) - 20,
        lambda x: (x - 12 if x >= 12 else 20 + x) << 3,
    ),
    "rpm_labor": (
        lambda x: x,
        lambda rpm, labor: ((labor - 12 if labor >= 12 else 20 + labor) << 3) | rpm & 0b111,
    ),
}

C = TypeVar("C", bound="CompData")
R = TypeVar("R", bound=CommandReq)


class CompData(ABC, Generic[R]):
    """
    CompData and it's subclasses are used to hold component state
    received from the Lionel Base 3 on Engines, Trains, Routes, Switches,
    and Accessories. Using the appropriate dict of CompDataHandler objects,
    encoded byte strings from the Base 3 are parsed into Python types. This
    information is used by ComponentState subclasses to maintain the current
    state of a specific component, as well as communicate that state to
    PyTrain clients from the server

    Each CompData subclass defines a set of fields that will contain the
    state information. By overriding __getattr__ and __setattr__, we can
    access the state as simple python properties as well as set them. Only
    those fields defined in the subclass are accessible; attempting to access
    others results in an AttributeError exception.
    """

    __metaclass__ = ABCMeta

    @classmethod
    def from_bytes(cls, data: bytes, scope: CommandScope, tmcc_id: int = None) -> C:
        """
        Parse byte packet into a CompData object, based on scope.
        """
        if scope == CommandScope.ENGINE:
            return EngineData(data, tmcc_id=tmcc_id)
        elif scope == CommandScope.TRAIN:
            return TrainData(data, tmcc_id=tmcc_id)
        elif scope == CommandScope.ACC:
            return AccessoryData(data, tmcc_id=tmcc_id)
        elif scope == CommandScope.SWITCH:
            return SwitchData(data, tmcc_id=tmcc_id)
        elif scope == CommandScope.ROUTE:
            return RouteData(data, tmcc_id=tmcc_id)
        else:
            raise ValueError(f"Invalid scope: {scope}")

    @classmethod
    def encode_rpm_labor(cls, rpm: int, labor: int) -> int:
        conv = CONVERSIONS.get("rpm_labor")
        return conv[1](rpm, labor)

    @classmethod
    def rpm_labor_to_pkg(cls, rpm: int, labor: int) -> UpdatePkg:
        return UpdatePkg("rpm_labor", 0xC0, 1, default_to_func(cls.encode_rpm_labor(rpm, labor)))

    # noinspection PyTypeChecker
    @classmethod
    def request_to_updates(cls, req: R) -> list[UpdatePkg] | None:
        """
        Pushes derivative state changes to Base 3 in response to a CommandReq.
        For example, when an engine is reset (Numeric 0), its speed and RPM
        must be reset to zero, and its labor must be set to 12. Derivative
        states are maintained in a dict keyed by the enum name corresponding
        to the CommandReq.
        """
        if not isinstance(req, CommandReq):
            raise AttributeError(f"'Argument is not a CommandReq: {req}'")

        update_pkgs: list[UpdatePkg] = []
        cmd = req.command
        updates = REQUEST_TO_UPDATES_MAP.get(cmd.name, None)
        if updates is None:
            return None
        for update in updates:
            if isinstance(update, tuple) and len(update) >= 1:
                field = sub_field = update[0]
                addr = FIELD_TO_ADDR_ENGINE_MAP.get(field, None)
                # special case for rpm/labor
                if addr is None and field in {"rpm", "labor"}:
                    field = "rpm_labor"
                    addr = FIELD_TO_ADDR_ENGINE_MAP.get(field, None)
                if addr is None:
                    log.warning(f"Field {field} not found in FIELD_TO_ADDR_ENGINE_MAP ({req})")
                    continue
                handler = BASE_MEMORY_ENGINE_READ_MAP.get(addr, None)
                if handler is None:
                    if addr is None:
                        log.warning(f"Field {field} handler not found in BASE_MEMORY_ENGINE_READ_MAP ({req})")
                    continue
                if len(update) == 1:
                    # have to convert data from command into Base 3 format
                    # is there a converter?
                    conv_tpl = CONVERSIONS.get(field, None)
                    if conv_tpl:
                        # more special case handling for rpm/labor
                        if sub_field != field and field == "rpm_labor":
                            from ..db.component_state_store import ComponentStateStore

                            state = ComponentStateStore.build().get_state(req.scope, req.address, False)
                            assert state is not None
                            with state.synchronizer:
                                if sub_field == "rpm":
                                    rpm = req.data
                                    labor = state.labor
                                else:
                                    rpm = state.rpm
                                    labor = req.data
                            base_value = conv_tpl[1](rpm, labor)
                        else:
                            base_value = conv_tpl[1](req.data)
                    else:
                        base_value = req.data
                else:
                    base_value = update[1](0)
                data_bytes = handler.to_bytes(base_value)
                if len(data_bytes) < handler.length:
                    data_bytes += b"\xff" * (handler.length - len(data_bytes))
                update_pkgs.append(UpdatePkg(field, addr, handler.length, data_bytes))
        return update_pkgs

    @abstractmethod
    def __init__(self, data: bytes, scope: CommandScope, tmcc_id: int = None) -> None:
        super().__init__()
        self._tmcc_id: int | None = tmcc_id
        self._scope = scope
        self._road_name_len: int | None = None
        self._road_name: str | None = None
        self._road_number_len: int | None = None
        self._road_number: str | None = None
        self._next_link: int | None = None
        self._prev_link: int | None = None

        # load the data from the byte string
        if data:
            self._parse_bytes(data, SCOPE_TO_COMP_MAP.get(self.scope))

    def __repr__(self) -> str:
        nm = nu = ""
        if self.road_name is not None:
            nm = f" {title(self.road_name)}"
        if self.road_number is not None:
            nu = f" #{self.road_number}"
        if self.scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
            return f"{self.scope.title} {self.tmcc_id:04}: {self.payload()}{nm}{nu}"
        else:
            return f"{self.scope.title} {self.tmcc_id:>2}: {self.payload()}{nm}{nu}"

    def __getattr__(self, name: str) -> Any:
        if name in self.__dict__:
            return self.__dict__[name]
        if "_" + name in self.__dict__:
            return self.__dict__["_" + name]
        elif name.endswith("_tmcc") and name.replace("_tmcc", "") in CONVERSIONS:
            name = name.replace("_tmcc", "")
            tpl = CONVERSIONS[name]
            # special case labor/rpm
            if name in {"rpm", "labor"}:
                name = "rpm_labor"
            value = self.__dict__["_" + name]
            return tpl[0](value) if value is not None else value
        else:
            raise AttributeError(f"'{type(self).__name__}' has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
            return
        if "_" + name in self.__dict__:
            self.__dict__["_" + name] = value
        elif name.endswith("_tmcc") and name.replace("_tmcc", "") in CONVERSIONS:
            name = name.replace("_tmcc", "")
            tpl = CONVERSIONS[name]
            # Special case rpm and labor, as they are encoded into a single value.
            # We determine the original values of each, then use our conversion
            # mechanism to use the conversion function to set the combined value.
            if name in {"rpm", "labor"}:
                rpm = self.rpm_tmcc if name == "labor" else value
                labor = self.labor_tmcc if name == "rpm" else value
                self.rpm_labor_tmcc = (rpm, labor)
            else:
                # For RPM or Labor, we have to pass the 2 raw values to the conversion function
                # as a tuple, thus requiring the isinstance check below.
                if isinstance(value, tuple):
                    self.__dict__["_" + name] = tpl[1](*value) if value is not None else value
                else:
                    self.__dict__["_" + name] = tpl[1](value) if value is not None else value
        else:
            raise AttributeError(f"'{type(self).__name__}' has no attribute '{name}'")

    def _signal_initializing(self) -> None:
        self.__dict__["__initializing__"] = True

    def __signal_initialized(self) -> None:
        self.__dict__["__initializing__"] = False

    def payload(self) -> str:
        return ""

    def as_bytes(self) -> bytes:
        comp_map = SCOPE_TO_COMP_MAP.get(self.scope)
        schema = {key: comp_map[key] for key in sorted(comp_map.keys())}
        if self.tmcc_id <= 99:
            # delete any entries that are 4-digit specific
            for k in list(schema.keys()):
                if schema[k].is_d4_only is True:
                    del schema[k]
        byte_str = bytes()
        last_idx = 0
        for idx, tpl in schema.items():
            if idx > last_idx:
                byte_str += b"\xff" * (idx - last_idx)
            data_len = tpl.length
            new_bytes = tpl.to_bytes(getattr(self, tpl.field))
            if len(new_bytes) < data_len:
                new_bytes += b"\xff" * (data_len - len(new_bytes))
            byte_str += new_bytes
            last_idx = idx + data_len
        # final check
        rec_len = PdiReq.scope_record_length(self.scope)
        if len(byte_str) < rec_len:
            byte_str += b"\xff" * (rec_len - len(byte_str))

        return byte_str

    def _parse_bytes(self, data: bytes, pmap: dict) -> None:
        if data is None:
            print(f"TMCC_ID: {self.tmcc_id} Scope: {self.scope}")
        data_len = len(data)
        for k, v in pmap.items():
            if isinstance(v, CompDataHandler) is False:
                continue
            item_len = v.length
            if data_len >= ((k + item_len) - 1) and hasattr(self, v.field) and getattr(self, v.field) is None:
                func = v.from_bytes
                try:
                    value = func(data[k : k + item_len])
                    if hasattr(self, v.field):
                        setattr(self, v.field, value)
                except Exception as e:
                    log.exception(f"Exception decoding {v.field} {e}", exc_info=e)


class EngineData(CompData):
    def __init__(
        self,
        data: bytes,
        tmcc_id: int = None,
        scope: CommandScope = CommandScope.ENGINE,
    ) -> None:
        self._signal_initializing()
        self._bt_id: int | None = None
        self._control_type: int | None = None
        self._engine_class: int | None = None
        self._engine_type: int | None = None
        self._max_speed: int | None = None
        self._momentum: int | None = None
        self._rpm_labor: int | None = None
        self._smoke: int | None = None
        self._sound_type: int | None = None
        self._speed: int | None = None
        self._speed_limit: int | None = None
        self._target_speed: int | None = None
        self._timestamp: int | None = None
        self._train_brake: int | None = None
        self._tsdb_left: int | None = None
        self._tsdb_right: int | None = None
        self._fuel_level: int | None = None
        self._water_level: int | None = None
        self._unk_2: int | None = None
        self._unk_3: int | None = None
        self._unk_6: int | None = None
        self._unk_a: int | None = None
        self._unk_b: int | None = None
        self._unk_f: int | None = None
        self._unk_11: int | None = None
        self._unk_12: int | None = None
        self._unk_13: int | None = None
        self._unk_17: int | None = None
        self._unk_5c: int | None = None
        self._unk_5e: int | None = None
        self._unk_68: int | None = None
        super().__init__(data, scope, tmcc_id=tmcc_id)


class TrainData(EngineData):
    """
    Represents train data within a Lionel layout.

    This class extends the functionality of `EngineData` to handle train-specific
    information. It is designed to process and manage data related to the train's
    operation, such as consist flags and consist components, in addition to providing
    integration capabilities with the engine control system.

    Attributes:
        _consist_flags (int | None): Flags representing the state or configuration of the
            train. This value is optional and may be set to `None` if not applicable.
        _consist_comps (list[ConsistComponent] | None): A list of components that make
            up the train's consist. This is optional and can remain `None` if no consist
            components are defined.
    """

    def __init__(self, data: bytes, tmcc_id: int = None) -> None:
        self._signal_initializing()
        self._consist_flags: int | None = None
        self._consist_comps: list[ConsistComponent] | None = None
        super().__init__(data, tmcc_id=tmcc_id, scope=CommandScope.TRAIN)


class SwitchData(CompData):
    """
    Represents the SwitchData class which extends the CompData class.

    This class is designed to handle and initialize switch-related data operations,
    using the base functionality provided by the CompData class. It incorporates
    specialized operations required for switch data management and processing. Primarily,
    it initializes TMCC ID, Name, Number, and state during its construction phase.
    """

    def __init__(self, data: bytes, tmcc_id: int = None) -> None:
        self._signal_initializing()
        super().__init__(data, scope=CommandScope.SWITCH, tmcc_id=tmcc_id)


class AccessoryData(CompData):
    def __init__(self, data: bytes, tmcc_id: int = None) -> None:
        self._signal_initializing()
        self._device_code: int | None = None
        super().__init__(data, scope=CommandScope.ACC, tmcc_id=tmcc_id)


class RouteData(CompData):
    def __init__(self, data: bytes, tmcc_id: int = None) -> None:
        self._signal_initializing()
        self._components: list[RouteComponent] | None = None
        super().__init__(data, scope=CommandScope.ROUTE, tmcc_id=tmcc_id)

    def payload(self) -> str:
        sw = ""
        if self._components:
            sw = "Switches: "
            sep = ""
            for c in self._components:
                state = "thru" if c.is_thru else "out"
                sw += f"{sep}{c.tmcc_id:>2} [{state}]"
                sep = ", "
        return sw


class CompDataMixin(Generic[C]):
    """
    Provides a mixin class for managing component-related data and
    recording state in generic types.

    This mixin class is designed to extend the functionality of a
    base class by adding attributes for component-related data
    and a recording state flag. The generic type `C` is used to
    allow flexibility in the type of component data stored.
    The mixin includes properties to access the component data
    and determine whether the recording state is set.

    Attributes:
        _comp_data: A generic component-related data attribute
            of type `C`. Defaults to None.
        _comp_data_record: A flag indicating whether component
            data recording is active. Defaults to False.
    """

    def __init__(self):
        super().__init__()
        self._comp_data: C | None = None
        self._comp_data_record: bool = False

    @property
    def comp_data(self) -> C:
        return self._comp_data

    @property
    def is_comp_data_record(self) -> bool:
        return self.comp_data and self._comp_data_record is True

    def initialize(self, scope: CommandScope, tmcc_id: int) -> None:
        data_len = PdiReq.scope_record_length(scope)
        if scope == CommandScope.SWITCH:
            # noinspection PyTypeChecker
            self._comp_data = SwitchData(b"\xff" * data_len, tmcc_id)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import socket
import subprocess
from multiprocessing import Pool, cpu_count
from typing import List


from ..protocol.constants import DEFAULT_BASE_PORT


def get_ip_address(max_attempts: int = 32) -> List[str]:
    from .. import is_linux

    # if on linux, use hostname to get IP addr
    if is_linux():
        result = subprocess.run("hostname -I".split(), capture_output=True, text=True)
        if result.returncode == 0:
            output = result.stdout.strip().split()
            if output and output[0]:
                return [output[0]]

    # Otherwise, use socket technique
    hostname = socket.gethostname()
    hostname = hostname if hostname.endswith(".local") else hostname + ".local"

    # Step 2: Get a list of IP addresses associated with the hostname
    ip_addresses = []
    attempts = 0
    while len(ip_addresses) == 0 and attempts <= max_attempts:
        try:
            ip_addresses = socket.gethostbyname_ex(hostname)[2]
        except socket.gaierror as ge:
            attempts += 1
            if attempts > max_attempts:
                raise ge
        except socket.herror as he:
            attempts += 1
            if attempts > max_attempts:
                raise he

    filtered_ips = {ip for ip in ip_addresses if not ip.startswith("127.")}
    return list(filtered_ips)


def is_base_address(address, base3_port: int = DEFAULT_BASE_PORT) -> str | None:
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(0.15)
            s.connect((address, base3_port))
            return address
    except socket.error:
        return None


def find_base_address() -> str | None:
    num_cpus = cpu_count() - 1
    possible_ips = list()
    local_ips = get_ip_address()
    for ip in local_ips:
        parts = ip.split(".")
        me = int(parts[-1])
        network = ".".join(parts[:-1]) + "."
        for i in range(2, 255):
            if i == me:
                continue
            possible_ips.append(network + str(i))
    with Pool(num_cpus) as p:
        for result in p.imap_unordered(is_base_address, possible_ips):
            if result is not None:
                p.terminate()
                return result
    return None



import random
from typing import TypeVar

from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.multibyte.multibyte_constants import *
from src.pytrain.protocol.tmcc1.tmcc1_constants import *
from src.pytrain.protocol.tmcc2.tmcc2_constants import *
from src.pytrain.utils.validations import Validations

T = TypeVar("T", TMCC1Enum, TMCC2Enum)


# noinspection PyMethodMayBeStatic
class TestBase:
    thread_exceptions = {}

    def teardown_method(self, test_method):
        pass

    def setup_method(self, test_method):
        pass

    def clear_thread_exceptions(self):
        self.thread_exceptions.clear()

    def custom_excepthook(self, args):
        new_exception = {
            "thread": args.thread,
            "exception": {"type": args.exc_type, "value": args.exc_value, "traceback": args.exc_traceback},
        }
        thread_name = args.thread.name
        if thread_name in self.thread_exceptions:
            self.thread_exceptions[thread_name].append(new_exception)
        else:
            self.thread_exceptions[thread_name] = [new_exception]

    def build_request(self, cmd, address: int = None, data: int = None, scope: CommandScope = None) -> CommandReq:
        if address is None:
            address = self.generate_random_address(cmd, scope=scope)
        if data is None:
            if cmd.value.is_data > 0:
                data = self.generate_random_data(cmd)
            else:
                data = 0
        return CommandReq.build(cmd, address, data, scope)

    @property
    def all_command_enums(self) -> List[T]:
        return [
            TMCC1HaltCommandEnum,
            TMCC1RouteCommandEnum,
            TMCC1AuxCommandEnum,
            TMCC1SwitchCommandEnum,
            TMCC1EngineCommandEnum,
            TMCC2HaltCommandEnum,
            TMCC2RouteCommandEnum,
            TMCC2EngineCommandEnum,
            TMCC2RailSoundsDialogControl,
            TMCC2EffectsControl,
            TMCC2LightingControl,
        ]

    def generate_random_address(self, cmd: CommandDefEnum, scope: CommandScope = None) -> int:
        if cmd.syntax == CommandSyntax.TMCC:
            if scope == CommandScope.TRAIN or cmd.scope == CommandScope.ROUTE:
                address = random.randint(1, 10)
            else:
                address = random.randint(1, 99)
        else:
            address = random.randint(1, 99)
        return address

    def generate_random_data(self, cmd: CommandDefEnum) -> int:
        command_def = cmd.value
        if command_def.data_max != 0:
            data = random.randint(command_def.data_min, command_def.data_max)
        elif command_def.data_map:
            data = random.randint(min(command_def.data_map), max(command_def.data_map))
        elif command_def.num_data_bits == 0:
            data = 0
        else:
            raise ValueError(f"Invalid command def: {cmd.name}")
        return data


# noinspection PyUnusedLocal
class MockCommandReq:
    @classmethod
    def _enqueue_command(
        cls,
        cmd: bytes,
        repeat: int,
        delay: float,
        baudrate: int,
        port: str,
        server: str,
        buffer,
        original_req: "MockCommandReq",
    ) -> bytes:
        repeat = Validations.validate_int(repeat, min_value=1, label="repeat")
        delay = Validations.validate_float(delay, min_value=0, label="delay")
        print(f"In MockCommandRequest._enqueue_command: {cmd.hex()} {repeat}, {delay}, {baudrate}")

        cmd_bytes: bytes = bytes()
        for _ in range(repeat):
            cmd_bytes += cmd
        return cmd_bytes


# noinspection PyMethodMayBeStatic
class MockCommBuffer:
    _instance = None

    def __init__(self, baudrate: int, port: str):
        if MockCommBuffer._instance.__initialized:
            return
        else:
            MockCommBuffer._instance.__initialized = True
        self.baudrate = baudrate
        self.port = port

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in the system
        """
        if not cls._instance:
            cls._instance = super(MockCommBuffer, cls).__new__(cls)
            cls._instance.__initialized = False
        return cls._instance

    def enqueue_command(self, command: bytes) -> bytes:
        return command



#!/usr/bin/env python3
#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import logging
from argparse import ArgumentParser
from typing import List

from ..protocol.tmcc1.acc_cmd import AccCmd
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from ..utils.argument_parser import PyTrainArgumentParser
from . import CliBase, DataAction

log = logging.getLogger(__name__)

AUX_OPTIONS_MAP = {
    "on": "ON",
    "off": "OFF",
    "opt1": "OPT_ONE",
    "opt2": "OPT_TWO",
}


class AccCli(CliBase):
    @classmethod
    def command_parser(cls) -> ArgumentParser:
        acc_parser = PyTrainArgumentParser(add_help=False)
        acc_parser.add_argument("acc", metavar="Accessory Number", type=int, help="accessory to fire")

        aux_group = acc_parser.add_mutually_exclusive_group()
        aux_group.add_argument(
            "-aux1",
            dest="aux1",
            choices=["on", "off", "opt1", "opt2"],
            nargs="?",
            type=str,
            const="opt1",
        )
        aux_group.add_argument(
            "-aux2",
            dest="aux2",
            choices=["on", "off", "opt1", "opt2"],
            nargs="?",
            type=str,
            const="opt1",
        )
        aux_group.add_argument(
            "-n",
            action=DataAction,
            dest="command",
            choices=range(0, 10),
            metavar="0 - 9",
            type=int,
            nargs="?",
            default=1,
            const=TMCC1AuxCommandEnum.NUMERIC,
            help="Numeric value",
        )
        aux_group.add_argument(
            "-speed",
            action=DataAction,
            dest="command",
            choices=range(-5, 6),
            metavar="-5 - 5",
            type=int,
            nargs="?",
            const=TMCC1AuxCommandEnum.RELATIVE_SPEED,
            help="Relative speed",
        )
        aux_group.add_argument(
            "-address",
            action="store_const",
            const=TMCC1AuxCommandEnum.SET_ADDRESS,
            dest="command",
            help="Set accessory address",
        )
        # fire command
        return PyTrainArgumentParser(
            "Operate specified accessory (1 - 99)",
            parents=[acc_parser, cls.cli_parser(), cls.multi_parser()],
        )

    """
        Issue Accessory Commands.

        Currently only available via the TMCC1 command format
    """

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        self._acc = self._args.acc
        self._command: TMCC1AuxCommandEnum | AccCmd = self._args.command
        self._data = self._args.data if "data" in self._args else 0
        self._speed = self._args.speed if "speed" in self._args else 0
        self._aux1 = self._args.aux1
        self._aux2 = self._args.aux2
        if self._args.aux1 and self._args.aux1 in AUX_OPTIONS_MAP:
            self._command = TMCC1AuxCommandEnum.by_name(f"AUX1_{AUX_OPTIONS_MAP[self._args.aux1]}")
        elif self._args.aux2 and self._args.aux2 in AUX_OPTIONS_MAP:
            self._command = TMCC1AuxCommandEnum.by_name(f"AUX2_{AUX_OPTIONS_MAP[self._args.aux2]}")
        try:
            if self._command is None or not isinstance(self._command, TMCC1AuxCommandEnum):
                raise ValueError("Must specify an option, use -h for help")
            cmd = AccCmd(
                self._acc,
                self._command,
                self._data,
                baudrate=self._baudrate,
                port=self._port,
                server=self._server,
            )
            if self.do_fire:
                cmd.fire(
                    repeat=self._args.repeat,
                    delay=self._args.delay,
                    baudrate=self._baudrate,
                    port=self._port,
                    server=self._server,
                )
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum
from ..utils.validations import Validations
from .gpio_device import GpioDevice, P
from .state_source import EngineStateSource


class CraneCar(GpioDevice):
    def __init__(
        self,
        address: int,
        cab_left_pin: P = None,
        cab_right_pin: P = None,
        bo_down_pin: P = None,
        bo_up_pin: P = None,
        bh_down_pin: P = None,
        bh_up_pin: P = None,
        sh_down_pin: P = None,
        sh_up_pin: P = None,
        bo_pin: P = None,
        bh_pin: P = None,
        sh_pin: P = None,
        fl_pin: P = None,
        rl_pin: P = None,
        or_pin: P = None,
        fc_pin: P = None,
        rc_pin: P = None,
        bo_led_pin: P = None,
        bh_led_pin: P = None,
        sh_led_pin: P = None,
        cathode: bool = True,
        cab_rotary_encoder: bool = False,
        repeat_every: float = 0.02,
    ) -> None:
        Validations.validate_float(repeat_every, 0.005, 1, label="repeat_every")
        if cab_rotary_encoder:
            from .py_rotary_encoder import PyRotaryEncoder

            self.cab_left_btn = self.cab_right_btn = None
            cmd = CommandReq.build(TMCC1EngineCommandEnum.RELATIVE_SPEED, address, data=0, scope=CommandScope.ENGINE)
            self.cab_re = PyRotaryEncoder(
                cab_left_pin,
                cab_right_pin,
                cmd,
                wrap=False,
                initial_step=0,
                max_steps=180,
                steps_to_data=self.std_step_to_data,
                pause_for=repeat_every,
                reset_after_motion=True,
            )
        else:
            # use momentary contact switch to rotate cab
            self.cab_re = None
            left_cmd, self.cab_left_btn, _ = self.make_button(
                cab_left_pin,
                command=TMCC1EngineCommandEnum.RELATIVE_SPEED,
                address=address,
                data=-1,
                scope=CommandScope.ENGINE,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.cab_left_btn.when_pressed = left_cmd.as_action()
            self.cab_left_btn.when_held = left_cmd.as_action()

            right_cmd, self.cab_right_btn, _ = self.make_button(
                cab_right_pin,
                command=TMCC1EngineCommandEnum.RELATIVE_SPEED,
                address=address,
                data=1,
                scope=CommandScope.ENGINE,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.cab_right_btn.when_pressed = right_cmd.as_action()
            self.cab_right_btn.when_held = right_cmd.as_action()

        # boom control
        boom_sel_cmd = CommandReq.build(TMCC1EngineCommandEnum.NUMERIC, address, data=1, scope=CommandScope.ENGINE)
        self.bo_btn = self.bo_led = None
        if bo_pin is not None:
            cmd, self.bo_btn, self.bo_led = self.when_button_pressed(
                bo_pin,
                boom_sel_cmd,
                led_pin=bo_led_pin,
                cathode=cathode,
            )
            if self.bo_led:
                self.cache_handler(EngineStateSource(address, self.bo_led, lambda x: x.numeric == 1))

        # large hook control
        bh_sel_cmd = CommandReq.build(TMCC1EngineCommandEnum.NUMERIC, address, data=2, scope=CommandScope.ENGINE)
        self.bh_btn = self.bh_led = None
        if bh_pin is not None:
            cmd, self.bh_btn, self.bh_led = self.when_button_pressed(
                bh_pin,
                bh_sel_cmd,
                led_pin=bh_led_pin,
                cathode=cathode,
            )
            if self.bh_led:
                self.cache_handler(EngineStateSource(address, self.bh_led, lambda x: x.numeric == 2))

        # small hook control
        sh_sel_cmd = CommandReq.build(TMCC1EngineCommandEnum.NUMERIC, address, data=3, scope=CommandScope.ENGINE)
        self.sh_btn = self.sh_led = None
        if sh_pin is not None:
            cmd, self.sh_btn, self.sh_led = self.when_button_pressed(
                sh_pin,
                sh_sel_cmd,
                led_pin=sh_led_pin,
                cathode=cathode,
            )
            if self.sh_led:
                self.cache_handler(EngineStateSource(address, self.sh_led, lambda x: x.numeric == 3))

        # set-up for boom lift/lower
        # we can either press a selector button to force crane mode or
        # send the prefix command along with the boom down command
        down_cmd = CommandReq.build(TMCC1EngineCommandEnum.BRAKE_SPEED, address)
        if bo_down_pin:
            _, self.down_btn, _ = self.make_button(
                bo_down_pin,
                down_cmd,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            if bh_down_pin or sh_down_pin:
                self.down_btn.when_pressed = self.with_prefix_action(boom_sel_cmd, down_cmd)
            else:
                self.down_btn.when_pressed = down_cmd.as_action()
            self.down_btn.when_held = down_cmd.as_action()
        else:
            self.down_btn = None

        up_cmd = CommandReq.build(TMCC1EngineCommandEnum.BOOST_SPEED, address)
        if bo_up_pin:
            _, self.up_btn, _ = self.make_button(
                bo_up_pin,
                up_cmd,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            if bh_up_pin or sh_up_pin:
                self.up_btn.when_pressed = self.with_prefix_action(boom_sel_cmd, up_cmd)
            else:
                self.up_btn.when_pressed = up_cmd.as_action()
            self.up_btn.when_held = up_cmd.as_action()
        else:
            self.up_btn = None

        # big hook up/down
        if bh_down_pin:
            _, self.bh_down_btn, _ = self.make_button(
                bh_down_pin,
                down_cmd,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            if bo_down_pin or sh_down_pin:
                self.bh_down_btn.when_pressed = self.with_prefix_action(bh_sel_cmd, down_cmd)
            else:
                self.bh_down_btn.when_pressed = down_cmd.as_action()
            self.bh_down_btn.when_held = down_cmd.as_action()
        else:
            self.bh_down_btn = None

        if bh_up_pin:
            _, self.bh_up_btn, _ = self.make_button(
                bh_up_pin,
                up_cmd,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            if bo_up_pin or sh_up_pin:
                self.bh_up_btn.when_pressed = self.with_prefix_action(bh_sel_cmd, up_cmd)
            else:
                self.bh_up_btn.when_pressed = up_cmd.as_action()
            self.bh_up_btn.when_held = up_cmd.as_action()
        else:
            self.bh_up_btn = None

        # small hook up/down
        if sh_down_pin:
            _, self.sh_down_btn, _ = self.make_button(
                sh_down_pin,
                down_cmd,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            if bo_down_pin or bh_down_pin:
                self.sh_down_btn.when_pressed = self.with_prefix_action(sh_sel_cmd, down_cmd)
            else:
                self.sh_down_btn.when_pressed = down_cmd.as_action()
            self.sh_down_btn.when_held = down_cmd.as_action()
        else:
            self.sh_down_btn = None

        if sh_up_pin:
            _, self.sh_up_btn, _ = self.make_button(
                sh_up_pin,
                up_cmd,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            if bo_up_pin or bh_up_pin:
                self.sh_up_btn.when_pressed = self.with_prefix_action(sh_sel_cmd, up_cmd)
            else:
                self.sh_up_btn.when_pressed = up_cmd.as_action()
            self.sh_up_btn.when_held = up_cmd.as_action()
        else:
            self.sh_up_btn = None

        # front/rear lights
        if fl_pin:
            fl_cmd, self.fl_btn, _ = self.make_button(
                fl_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=4,
                scope=CommandScope.ENGINE,
            )
            self.fl_btn.when_pressed = fl_cmd.as_action()
        else:
            self.fl_btn = None

        if rl_pin:
            rl_cmd, self.rl_btn, _ = self.make_button(
                rl_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=5,
                scope=CommandScope.ENGINE,
            )
            self.rl_btn.when_pressed = rl_cmd.as_action()
        else:
            self.rl_btn = None

        if or_pin:
            or_cmd, self.or_btn, _ = self.make_button(
                or_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=6,
                scope=CommandScope.ENGINE,
            )
            self.or_btn.when_pressed = or_cmd.as_action()
        else:
            self.or_btn = None

        # front/rear coupler
        if fc_pin:
            fc_cmd, self.fc_btn, _ = self.make_button(
                fc_pin,
                command=TMCC1EngineCommandEnum.FRONT_COUPLER,
                address=address,
                scope=CommandScope.ENGINE,
            )
            self.fc_btn.when_pressed = fc_cmd.as_action()
        else:
            self.fc_btn = None

        if rc_pin:
            rc_cmd, self.rc_btn, _ = self.make_button(
                rc_pin,
                command=TMCC1EngineCommandEnum.REAR_COUPLER,
                address=address,
                scope=CommandScope.ENGINE,
            )
            self.rc_btn.when_pressed = rc_cmd.as_action()
        else:
            self.rc_btn = None



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from typing import Dict, Any, cast

from .comp_data import CompDataMixin
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum as Aux, TMCC1HaltCommandEnum
from ..protocol.command_req import CommandReq
from ..pdi.asc2_req import Asc2Req
from ..pdi.bpc2_req import Bpc2Req
from ..pdi.constants import Asc2Action, Bpc2Action, PdiCommand, IrdaAction
from ..pdi.irda_req import IrdaReq
from .component_state import TmccState, L, P, log, SCOPE_TO_STATE_MAP


class AccessoryState(TmccState):
    def __init__(self, scope: CommandScope = CommandScope.ACC) -> None:
        if scope != CommandScope.ACC:
            raise ValueError(f"Invalid scope: {scope}")
        super().__init__(scope)
        self._first_pdi_command = None
        self._first_pdi_action = None
        self._last_aux1_opt1 = None
        self._last_aux2_opt1 = None
        self._aux1_state: Aux | None = None
        self._aux2_state: Aux | None = None
        self._aux_state: Aux | None = None
        self._block_power = False
        self._sensor_track = False
        self._pdi_source = False
        self._number: int | None = None

    @property
    def payload(self) -> str:
        aux1 = aux2 = aux_num = ""
        if self._block_power:
            aux = f"Block Power {'ON' if self.aux_state == Aux.AUX1_OPT_ONE else 'OFF'}"
        elif self._sensor_track:
            aux = "Sensor Track"
        else:
            if self.is_lcs_component:
                aux = "Asc2 " + "ON" if self._aux_state == Aux.AUX1_OPT_ONE else "OFF"
            else:
                if self.aux_state == Aux.AUX1_OPT_ONE:
                    aux = "Aux 1"
                elif self.aux_state == Aux.AUX2_OPT_ONE:
                    aux = "Aux 2"
                else:
                    aux = "Unknown"
                aux1 = f" Aux1: {self.aux1_state.name if self.aux1_state is not None else 'Unknown'}"
                aux2 = f" Aux2: {self.aux2_state.name if self.aux2_state is not None else 'Unknown'}"
                aux_num = f" Aux Num: {self._number if self._number is not None else 'NA'}"
        return f"{aux}{aux1}{aux2}{aux_num}"

    # noinspection DuplicatedCode
    def update(self, command: L | P) -> None:
        if command:
            with self._cv:
                super().update(command)
                if isinstance(command, CompDataMixin) and command.is_comp_data_record:
                    self._update_comp_data(command.comp_data)
                elif isinstance(command, CommandReq):
                    if command.command != Aux.SET_ADDRESS:
                        if command.command == TMCC1HaltCommandEnum.HALT:
                            self._aux1_state = Aux.AUX1_OFF
                            self._aux2_state = Aux.AUX2_OFF
                            self._aux_state = Aux.AUX2_OPT_ONE
                            self._number = None
                        else:
                            if self._pdi_source is False:
                                if command.command in {Aux.AUX1_OPT_ONE, Aux.AUX2_OPT_ONE}:
                                    self._aux_state = command.command
                                if command.command == Aux.AUX1_OPT_ONE:
                                    if self.time_delta(self._last_updated, self._last_aux1_opt1) > 1:
                                        self._aux1_state = self.update_aux_state(
                                            self._aux1_state,
                                            Aux.AUX1_ON,
                                            Aux.AUX1_OPT_ONE,
                                            Aux.AUX1_OFF,
                                        )
                                    self._last_aux1_opt1 = self.last_updated
                                elif command.command in {Aux.AUX1_ON, Aux.AUX1_OFF, Aux.AUX1_OPT_TWO}:
                                    self._aux1_state = command.command
                                    self._last_aux1_opt1 = self.last_updated
                                elif command.command == Aux.AUX2_OPT_ONE:
                                    if self.time_delta(self._last_updated, self._last_aux2_opt1) > 1:
                                        self._aux2_state = self.update_aux_state(
                                            self._aux2_state,
                                            Aux.AUX2_ON,
                                            Aux.AUX2_OPT_ONE,
                                            Aux.AUX2_OFF,
                                        )
                                    self._last_aux2_opt1 = self.last_updated
                                elif command.command in {Aux.AUX2_ON, Aux.AUX2_OFF, Aux.AUX2_OPT_TWO}:
                                    self._aux2_state = command.command
                                    self._last_aux2_opt1 = self.last_updated
                            if command.command == Aux.NUMERIC:
                                self._number = command.data
                elif isinstance(command, Asc2Req) or isinstance(command, Bpc2Req):
                    if self._first_pdi_command is None:
                        self._first_pdi_command = command.command
                    if self._first_pdi_action is None:
                        self._first_pdi_action = command.action
                    if command.action in {Asc2Action.CONTROL1, Bpc2Action.CONTROL1, Bpc2Action.CONTROL3}:
                        self._pdi_source = True
                        if command.action in {Bpc2Action.CONTROL1, Bpc2Action.CONTROL3}:
                            self._block_power = True
                        else:
                            self._block_power = False
                        if command.state == 1:
                            self._aux1_state = Aux.AUX1_ON
                            self._aux2_state = Aux.AUX2_ON
                            self._aux_state = Aux.AUX1_OPT_ONE
                        else:
                            self._aux1_state = Aux.AUX1_OFF
                            self._aux2_state = Aux.AUX2_OFF
                            self._aux_state = Aux.AUX2_OPT_ONE
                elif isinstance(command, IrdaReq):
                    if self._first_pdi_command is None:
                        self._first_pdi_command = command.command
                    if self._first_pdi_action is None:
                        self._first_pdi_action = command.action
                    self._sensor_track = True
                self.changed.set()
                self._cv.notify_all()

    @property
    def is_known(self) -> bool:
        return (
            self._aux_state is not None
            or self._aux1_state is not None
            or self._aux2_state is not None
            or self._number is not None
        )

    @property
    def is_power_district(self) -> bool:
        return self._block_power

    @property
    def is_sensor_track(self) -> bool:
        return self._sensor_track

    @property
    def is_lcs_component(self) -> bool:
        return self._pdi_source

    @property
    def aux_state(self) -> Aux:
        return self._aux_state

    @property
    def is_aux_on(self) -> bool:
        return self._aux_state == Aux.AUX1_OPT_ONE

    @property
    def is_aux_off(self) -> bool:
        return self._aux_state == Aux.AUX2_OPT_ONE

    @property
    def aux1_state(self) -> Aux:
        return self._aux1_state

    @property
    def aux2_state(self) -> Aux:
        return self._aux2_state

    @property
    def value(self) -> int:
        return self._number

    def as_bytes(self) -> bytes:
        if self.comp_data is None:
            self.initialize(self.scope, self.address)
        byte_str = super().as_bytes()
        if self._sensor_track:
            byte_str += IrdaReq(self.address, PdiCommand.IRDA_RX, IrdaAction.INFO, scope=CommandScope.ACC).as_bytes
        elif self.is_lcs_component:
            if isinstance(self._first_pdi_action, Asc2Action):
                byte_str += Asc2Req(
                    self.address,
                    self._first_pdi_command,
                    cast(Asc2Action, self._first_pdi_action),
                    values=1 if self._aux_state == Aux.AUX1_OPT_ONE else 0,
                ).as_bytes
            elif isinstance(self._first_pdi_action, Bpc2Action):
                byte_str += Bpc2Req(
                    self.address,
                    self._first_pdi_command,
                    cast(Bpc2Action, self._first_pdi_action),
                    state=1 if self._aux_state == Aux.AUX1_OPT_ONE else 0,
                ).as_bytes
            else:
                log.error(f"State req for lcs device: {self._first_pdi_command.name} {self._first_pdi_action.name}")
        else:
            if self._aux_state is not None:
                byte_str += CommandReq.build(self.aux_state, self.address).as_bytes
            if self._aux1_state is not None:
                byte_str += CommandReq.build(self.aux1_state, self.address).as_bytes
            if self._aux2_state is not None:
                byte_str += CommandReq.build(self.aux2_state, self.address).as_bytes
        return byte_str

    def as_dict(self) -> Dict[str, Any]:
        d = super()._as_dict()
        if self._sensor_track:
            d["type"] = "sensor track"
        elif self._block_power:
            d["type"] = "power district"
            d["block"] = "on" if self._aux_state == Aux.AUX1_OPT_ONE else "off"
        else:
            d["type"] = "accessory"
            d["aux"] = self._aux_state.name.lower() if self._aux_state else None
            d["aux1"] = self.aux1_state.name.lower() if self.aux1_state else None
            d["aux2"] = self.aux2_state.name.lower() if self.aux2_state else None
        return d


SCOPE_TO_STATE_MAP.update({CommandScope.ACC: AccessoryState})



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from ..db.component_state_store import ComponentStateStore
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1RouteCommandEnum
from .gpio_device import GpioDevice, P
from .state_source import RouteStateSource


class Route(GpioDevice):
    """
    Class Route is responsible for controlling and firing TMCC2/Legacy routes by managing
    the associated turnouts and using GPIO buttons and LED indications if available.

    This class handles the functionality of throwing all turnouts that are part of a
    specified route to their correct states. It binds actions to GPIO buttons for user
    interaction and optionally controls LED indicators to reflect the route's state.

    Attributes:
        route_btn (Button): Represents the button used to fire the route.
        active_led (Optional[LED]): Represents the LED associated with the route if provided.
        inactive_led (Optional[LED]): Represents the LED associated with the route if provided.

    Args:
        address (int): The unique TMCC_ID for the route to be fired.
        btn_pin (P): GPIO pin to which the physical button is connected.
        active_pin (Optional[P]): GPIO pin for a physical LED indicator, if any. Defaults to None.
        inactive_pin (Optional[P]): GPIO pin for a physical LED indicator, if any. Defaults to None.
        cathode (bool): Specifies the cathode-state configuration for the LED. Defaults to True.
    """

    def __init__(
        self,
        address: int,
        btn_pin: P,
        active_pin: P = None,
        inactive_pin: P = None,
        cathode: bool = True,
    ) -> None:
        """
        Fire a TMCC2/Legacy Route, throwing all incorporated turnouts to the correct state
        """
        self.state = ComponentStateStore.get_state(CommandScope.ROUTE, address, create=False)
        assert self.state

        # make the CommandReq
        req, self.route_btn, self.active_led = self.make_button(
            btn_pin,
            TMCC1RouteCommandEnum.FIRE,
            address,
            led_pin=active_pin,
            bind=False,
            initially_on=self.state.is_active,
            cathode=cathode,
        )

        if inactive_pin:
            self.inactive_led = self.make_led(inactive_pin, initially_on=not self.state.is_active, cathode=cathode)
        else:
            self.inactive_led = None

        # bind actions to buttons
        self.route_btn.when_pressed = req.as_action(repeat=2)

        if self.active_led:
            self.cache_handler(RouteStateSource(address, self.active_led, self.inactive_led))
